@@ -39,7 +39,7 @@ if TYPE_CHECKING:
 
 __all__ = ['MIN_TICKET_SIZE', 'CryptoError', 'OTPLengthError', 'CorruptBootromError', 'KeyslotMissingError',
            'TicketLengthError', 'BootromNotFoundError', 'CorruptOTPError', 'Keyslot', 'CryptoEngine', 'CTRFileIO',
-           'CBCFileIO']
+           'TWLCTRFileIO', 'CBCFileIO']
 
 BOOT9_PROT_HASH = '7331f7edece3dd33f2ab4bd0b3a5d607229fd19212c10b734cedcaf78c1a7b98'
 
@@ -161,6 +161,12 @@ class Keyslot(IntEnum):
     New3DSKeySector = 0x43
     """Used to decrypt the secret key sector (sector 0x96) for the New Nintendo 3DS."""
 
+    NCCHExtraKey = 0x44
+    """
+    Stores a version of another keyslot used for NCCH titles. For titles without a seed, KeyY is taken from the NCCH
+    header. For titles with a seed, KeyY is seeded. KeyX is always the same as the source keyslot.
+    """
+
 
 common_key_y = (
     # eShop
@@ -499,9 +505,9 @@ class CryptoEngine:
             title_id = bytes.fromhex(title_id)
 
         if self.dev and common_key_index == 0:
-            self.set_normal_key(0x3D, DEV_COMMON_KEY_0)
+            self.set_normal_key(Keyslot.CommonKey, DEV_COMMON_KEY_0)
         else:
-            self.set_keyslot('y', 0x3D, common_key_y[common_key_index])
+            self.set_keyslot('y', Keyslot.CommonKey, common_key_y[common_key_index])
 
         cipher = self.create_cbc_cipher(Keyslot.CommonKey, title_id + (b'\0' * 8))
         self.set_normal_key(Keyslot.DecryptedTitlekey, cipher.decrypt(titlekey))
@@ -869,7 +875,7 @@ class CTRFileIO(_CryptoFileBase):
                 f'closefd={self._closefd!r})')
 
     def __hash__(self):
-        return hash((self._reader, self._keyslot, self._counter))
+        return hash((self._reader, self._keyslot, self._counter, id(self)))
 
     @_raise_if_file_closed
     def read(self, size: int = -1) -> bytes:
@@ -948,7 +954,7 @@ class CBCFileIO(_CryptoFileBase):
                 f'closefd={self._closefd!r})')
 
     def __hash__(self):
-        return hash((self._reader, self._keyslot, self._iv))
+        return hash((self._reader, self._keyslot, self._iv, id(self)))
 
     @_raise_if_file_closed
     def read(self, size: int = -1):
